"""System Bridge: GUI"""
import asyncio
from logging import Logger
import subprocess
import sys
from threading import Thread

from systembridgeshared.exceptions import ConnectionErrorException
from systembridgeshared.settings import Settings
from systembridgeshared.websocket_client import WebSocketClient


class GUIAttemptsExceededException(BaseException):
    """Raise this when the GUI attempts to start more too many times."""


async def start_gui(  # pylint: disable=keyword-arg-before-vararg
    logger: Logger,
    settings: Settings,
    attempt: int = 1,
    command: str = "main",
    *args,
) -> None:
    """Start the GUI"""
    if attempt > 2:
        logger.error("Failed to start GUI after 2 attempts")
        raise GUIAttemptsExceededException("Failed to start GUI after 3 attempts")
    if command == "main":
        logger.info(
            "Test WebSocket connection before starting GUI. Attempt #%s", attempt
        )
        websocket_client = WebSocketClient(settings)
        try:
            await websocket_client.connect()
            await websocket_client.close()
        except ConnectionErrorException:
            logger.warning("Could not connect to WebSocket. Retrying in 5 seconds")
            await asyncio.sleep(5)
            await start_gui(
                logger,
                settings,
                attempt + 1,
                command,
                *args,
            )
            return

    pgm_args = [
        sys.executable,
        "-m",
        "systembridgegui",
        command,
        *args,
    ]

    logger.info("Starting GUI: %s", pgm_args)
    with subprocess.Popen(pgm_args) as process:
        logger.info("GUI started with PID: %s", process.pid)
        if (exit_code := process.wait()) != 0:
            logger.error("GUI exited with code: %s", exit_code)
            await start_gui(
                logger,
                settings,
                attempt + 1,
                command,
                *args,
            )
            return
        logger.info("GUI exited with code: %s", exit_code)


def start_gui_sync(  # pylint: disable=keyword-arg-before-vararg
    logger: Logger,
    settings: Settings,
    command: str = "main",
    *args,
) -> None:
    """Start the GUI in a synchronous thread"""
    asyncio.run(
        start_gui(
            logger,
            settings,
            1,
            command,
            *args,
        )
    )


def start_gui_threaded(  # pylint: disable=keyword-arg-before-vararg
    logger: Logger,
    settings: Settings,
    command: str = "main",
    *args,
) -> None:
    """Start the GUI in a thread"""
    thread = Thread(
        target=start_gui_sync,
        args=(
            logger,
            settings,
            command,
            *args,
        ),
    )
    thread.start()



"""System Bridge: GPU"""
from __future__ import annotations

from typing import Optional

from systembridgeshared.base import Base
from systembridgeshared.common import make_key
from systembridgeshared.const import (
    COLUMN_HARDWARE_NAME,
    COLUMN_HARDWARE_TYPE,
    COLUMN_KEY,
    COLUMN_NAME,
    COLUMN_TYPE,
    COLUMN_VALUE,
)
from systembridgeshared.database import Database


class GPU(Base):
    """GPU"""

    def get_gpus(
        self,
        database: Database,
    ) -> list[str]:
        """Get GPUs"""
        gpus = []
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "gpu" in item[COLUMN_HARDWARE_TYPE].lower()
                and item[COLUMN_HARDWARE_NAME] not in gpus
            ):
                gpus.append(item[COLUMN_HARDWARE_NAME])
        return gpus

    def core_clock(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU core clock"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "gpu" in item[COLUMN_HARDWARE_TYPE].lower()
                and "clock" in item[COLUMN_TYPE].lower()
                and "core" in item[COLUMN_NAME].lower()
                and make_key(item[COLUMN_HARDWARE_NAME]) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU core clock: %s = %s",
                    item[COLUMN_KEY],
                    item[COLUMN_VALUE],
                )
                return item[COLUMN_VALUE]
        return None

    def core_load(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU core load"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "gpu" in item[COLUMN_HARDWARE_TYPE].lower()
                and "load" in item[COLUMN_TYPE].lower()
                and "core" in item[COLUMN_NAME].lower()
                and make_key(item[COLUMN_HARDWARE_NAME]) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU core load: %s = %s", item[COLUMN_KEY], item[COLUMN_VALUE]
                )
                return item[COLUMN_VALUE]
        return None

    def fan_speed(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU fan speed"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "gpu" in item[COLUMN_HARDWARE_TYPE].lower()
                and "fan" in item[COLUMN_TYPE].lower()
                and make_key(item[COLUMN_HARDWARE_NAME]) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU fan speed: %s = %s", item[COLUMN_KEY], item[COLUMN_VALUE]
                )
                return item[COLUMN_VALUE]
        return None

    def memory_clock(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU memory clock"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "gpu" in item[COLUMN_HARDWARE_TYPE].lower()
                and "clock" in item[COLUMN_TYPE].lower()
                and "memory" in item[COLUMN_NAME].lower()
                and make_key(item[COLUMN_HARDWARE_NAME]) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU memory clock: %s = %s",
                    item[COLUMN_KEY],
                    item[COLUMN_VALUE],
                )
                return item[COLUMN_VALUE]
        return None

    def memory_load(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU memory load"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "gpu" in item[COLUMN_HARDWARE_TYPE].lower()
                and "load" in item[COLUMN_TYPE].lower()
                and "memory" in item[COLUMN_NAME].lower()
                and make_key(item[COLUMN_HARDWARE_NAME]) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU memory load: %s = %s",
                    item[COLUMN_KEY],
                    item[COLUMN_VALUE],
                )
                return item[COLUMN_VALUE]
        return None

    def memory_free(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU memory free"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "gpu" in item[COLUMN_HARDWARE_TYPE].lower()
                and "memory" in item[COLUMN_NAME].lower()
                and "free" in item[COLUMN_NAME].lower()
                and make_key(item[COLUMN_HARDWARE_NAME]) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU memory free: %s = %s",
                    item[COLUMN_KEY],
                    item[COLUMN_VALUE],
                )
                return item[COLUMN_VALUE]
        return None

    def memory_used(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU memory used"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "gpu" in item[COLUMN_HARDWARE_TYPE].lower()
                and "memory" in item[COLUMN_NAME].lower()
                and "used" in item[COLUMN_NAME].lower()
                and make_key(item[COLUMN_HARDWARE_NAME]) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU memory used: %s = %s",
                    item[COLUMN_KEY],
                    item[COLUMN_VALUE],
                )
                return item[COLUMN_VALUE]
        return None

    def memory_total(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU memory total"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "gpu" in item[COLUMN_HARDWARE_TYPE].lower()
                and "memory" in item[COLUMN_NAME].lower()
                and "total" in item[COLUMN_NAME].lower()
                and make_key(item[COLUMN_HARDWARE_NAME]) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU memory total: %s = %s",
                    item[COLUMN_KEY],
                    item[COLUMN_VALUE],
                )
                return item[COLUMN_VALUE]
        return None

    def power(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU power usage"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "gpu" in item[COLUMN_HARDWARE_TYPE].lower()
                and "power" in item[COLUMN_TYPE].lower()
                and make_key(item[COLUMN_HARDWARE_NAME]) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU power: %s = %s", item[COLUMN_KEY], item[COLUMN_VALUE]
                )
                return item[COLUMN_VALUE]
        return None

    def temperature(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU temperature"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "gpu" in item[COLUMN_HARDWARE_TYPE].lower()
                and "temperature" in item[COLUMN_TYPE].lower()
                and "core" in item[COLUMN_NAME].lower()
                and make_key(item[COLUMN_HARDWARE_NAME]) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU temperature: %s = %s",
                    item[COLUMN_KEY],
                    item[COLUMN_VALUE],
                )
                return item[COLUMN_VALUE]
        return None



"""System Bridge: CPU"""
from __future__ import annotations

from typing import Optional

from psutil import (
    cpu_count,
    cpu_freq,
    cpu_percent,
    cpu_stats,
    cpu_times,
    cpu_times_percent,
    getloadavg,
)
from psutil._common import pcputimes, scpufreq, scpustats
from systembridgeshared.base import Base
from systembridgeshared.const import (
    COLUMN_HARDWARE_TYPE,
    COLUMN_KEY,
    COLUMN_TYPE,
    COLUMN_VALUE,
)
from systembridgeshared.database import Database


class CPU(Base):
    """CPU"""

    def count(self) -> int:
        """CPU count"""
        return cpu_count()

    def freq(self) -> scpufreq:
        """CPU frequency"""
        return cpu_freq()

    def freq_per_cpu(self) -> list[scpufreq]:  # pylint: disable=unsubscriptable-object
        """CPU frequency per CPU"""
        return cpu_freq(percpu=True)  # type: ignore

    def load_average(
        self,
    ) -> tuple[float, float, float]:  # pylint: disable=unsubscriptable-object
        """Get load average"""
        return getloadavg()

    def stats(self) -> scpustats:
        """CPU stats"""
        return cpu_stats()

    def temperature(
        self,
        database: Database,
    ) -> Optional[float]:
        """CPU temperature"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if item[COLUMN_HARDWARE_TYPE] is not None and (
                (
                    "cpu" in item[COLUMN_HARDWARE_TYPE].lower()
                    and "temperature" in item[COLUMN_TYPE].lower()
                )
                or (
                    "k10temp" in item[COLUMN_HARDWARE_TYPE].lower()
                    and "current" in item[COLUMN_TYPE].lower()
                )
            ):
                self._logger.debug(
                    "Found CPU temperature: %s = %s",
                    item[COLUMN_KEY],
                    item[COLUMN_VALUE],
                )
                return item[COLUMN_VALUE]
        return None

    def times(self) -> pcputimes:
        """CPU times"""
        return cpu_times(percpu=False)

    def times_percent(self) -> pcputimes:
        """CPU times percent"""
        return cpu_times_percent(interval=1, percpu=False)

    def times_per_cpu(
        self,
    ) -> list[pcputimes]:  # pylint: disable=unsubscriptable-object
        """CPU times per CPU"""
        return cpu_times(percpu=True)

    def times_per_cpu_percent(
        self,
    ) -> list[pcputimes]:  # pylint: disable=unsubscriptable-object
        """CPU times per CPU percent"""
        return cpu_times_percent(interval=1, percpu=True)

    def usage(self) -> float:
        """CPU usage"""
        return cpu_percent(interval=1, percpu=False)

    def usage_per_cpu(self) -> list[float]:  # pylint: disable=unsubscriptable-object
        """CPU usage per CPU"""
        return cpu_percent(interval=1, percpu=True)  # type: ignore

    def voltage(
        self,
        database: Database,
    ) -> Optional[float]:
        """CPU voltage"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "cpu" in item[COLUMN_HARDWARE_TYPE].lower()
                and "voltage" in item[COLUMN_TYPE].lower()
            ):
                self._logger.debug(
                    "Found CPU voltage: %s = %s", item[COLUMN_KEY], item[COLUMN_VALUE]
                )
                return item[COLUMN_VALUE]
        return None



"""System Bridge: Autostart Linux"""
import os
import sys

desktop_entry = f"""
[Desktop Entry]
Name=System Bridge
Comment=System Bridge
Exec={sys.executable} -m systembridgebackend
Icon={os.path.join(os.path.dirname(__file__),'../icon.png')}
Terminal=false
Type=Application
Categories=Application;
"""


def autostart_linux_disable():
    """Disable autostart for Linux"""
    path = os.path.expanduser("~/.config/autostart/systembridge.desktop")
    if os.path.exists(path):
        os.remove(path)


def autostart_linux_enable():
    """Enable autostart for Linux"""
    path = os.path.expanduser("~/.config/autostart/systembridge.desktop")
    if not os.path.exists(path):
        os.makedirs(path)
    path = os.path.join(path, "systembridge.desktop")
    if not os.path.exists(path):
        with open(path, "w", encoding="utf-8") as file:
            file.write(desktop_entry)



"""System Bridge: Module Listeners"""

import asyncio
from collections.abc import Callable
from collections.abc import Awaitable

from systembridgeshared.base import Base
from systembridgeshared.const import MODEL_MAP
from systembridgeshared.database import Database


class Listener:
    """Listener"""

    def __init__(
        self,
        listener_id: str,
        data_changed_callback: Callable[[str, dict], Awaitable[None]],
        modules: list[str],  # pylint: disable=unsubscriptable-object
    ) -> None:
        """Initialize"""
        self.id = listener_id
        self.data_changed_callback = data_changed_callback
        self.modules = modules


class Listeners(Base):
    """Module Listeners"""

    def __init__(
        self,
        database: Database,
        implemented_modules: list[str],  # pylint: disable=unsubscriptable-object
    ) -> None:
        """Initialize"""
        super().__init__()
        self._database = database
        self._implemented_modules = implemented_modules
        self._data: dict = {module: {} for module in self._implemented_modules}
        # pylint: disable=unsubscriptable-object
        self._registered_listeners: list[Listener] = []

    async def add_listener(
        self,
        listener_id: str,
        data_changed_callback: Callable[[str, dict], Awaitable[None]],
        modules: list[str],  # pylint: disable=unsubscriptable-object
    ) -> bool:
        """Add modules to listener"""
        for listner in self._registered_listeners:
            if listner.id == listener_id:
                self._logger.warning("Listener already registered: %s", listener_id)
                return True

        self._registered_listeners.append(
            Listener(listener_id, data_changed_callback, modules)
        )
        self._logger.info("Added listener: %s", listener_id)

        await self.refresh_data()
        return False

    async def refresh_data(self) -> None:
        """Refresh data"""
        if not self._database.connected:
            self._database.connect()

        # Get modules from registered listeners
        modules = []
        for listener in self._registered_listeners:
            for module in listener.modules:
                if module not in modules:
                    modules.append(module)

        self._logger.info("Refresh data: %s", modules)

        # Refresh data for each module
        tasks = [self.refresh_data_by_module(module) for module in modules]
        await asyncio.gather(*tasks)

        self._logger.info("Finished refreshing data: %s", modules)

    async def refresh_data_by_module(
        self,
        module: str,
    ) -> None:
        """Refresh data by module"""
        self._logger.info("Refresh data by module: %s", module)
        if module not in self._implemented_modules:
            self._logger.warning("Module to refresh not implemented: %s", module)
            return

        if (model := MODEL_MAP.get(module)) is None:
            self._logger.warning("Unknown model: %s", module)
            return

        new_data = self._database.table_data_to_ordered_dict(module)
        if new_data is None:
            self._logger.warning("No data found for module: %s", module)
            return

        if new_data != self._data[module]:
            self._logger.info("Data changed for module: %s", module)
            self._data[module] = model(**new_data).dict()
            for listener in self._registered_listeners:
                self._logger.info("Listener: %s - %s", listener.id, listener.modules)
                if module in listener.modules:
                    self._logger.info(
                        "Sending '%s' data to listener: %s", module, listener.id
                    )
                    await listener.data_changed_callback(module, new_data)

    def remove_all_listeners(self) -> None:
        """Remove all listeners"""
        self._registered_listeners.clear()

    def remove_listener(
        self,
        listener_id: str,
    ) -> bool:
        """Remove listener"""
        for listener in self._registered_listeners:
            if listener.id == listener_id:
                self._registered_listeners.remove(listener)
                self._logger.info("Removed listener: %s", listener_id)
                return True

        self._logger.info("Listener not found: %s", listener_id)
        return False



"""System Bridge: Server Authentication"""
from collections.abc import Callable
from functools import wraps

from sanic import Sanic, exceptions
from sanic.request import Request
from systembridgeshared.base import Base


class ApiKeyAuthentication(Base):
    """ApiKey Authentication"""

    def __init__(
        self,
        app=None,
        arg="key",
        header=None,
        keys=None,
    ):
        """Initialize"""
        super().__init__()
        self.api_keys = keys
        self.header = header
        self.arg = arg
        self.error = "Authentication required"
        self.token = None

        if app is not None:
            self.init_app(app)

    def init_app(
        self,
        app: Sanic,
    ):
        """Initialize"""
        app.register_middleware(self.open_session, "request")

    async def open_session(self, request):
        """Open session"""
        pass  # pylint: disable=unnecessary-pass

    async def _is_api_key(self, request):
        """Check key is valid api key"""
        if not (token := request.headers.get(self.header, None)):
            token = request.args.get(self.arg, None)

        self.token = token
        return token in self.api_keys

    def get_token(self):
        """Get logged api key"""
        return self.token if self.token else ""

    def add_key(self, key: str):
        """Add API Key"""
        self.api_keys.append(key)

    def key_required(
        self,
        handler: Callable,
    ):
        """Wrap handler with key check"""

        @wraps(handler)
        async def wrapper(
            request: Request,
            *args,
            **kwargs,
        ) -> Callable:
            """Wrap handler with key check"""
            if not await self._is_api_key(request):
                raise exceptions.Unauthorized(self.error)
            return await handler(request, *args, **kwargs)

        return wrapper



"""System Bridge: Update Display"""
import asyncio

from systembridgeshared.common import make_key
from systembridgeshared.database import Database

from systembridgebackend.modules.base import ModuleUpdateBase
from systembridgebackend.modules.display import Display


class DisplayUpdate(ModuleUpdateBase):
    """Display Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database, "display")
        self._display = Display()

    async def update_name(
        self,
        display_key: str,
        display_name: str,
    ) -> None:
        """Update name"""
        self._database.write("display", f"{display_key}_name", display_name)

    async def update_pixel_clock(
        self,
        display_key: str,
    ) -> None:
        """Update pixel clock"""
        self._database.write(
            "display",
            f"{display_key}_pixel_clock",
            self._display.pixel_clock(self._database, display_key),
        )

    async def update_refresh_rate(
        self,
        display_key: str,
    ) -> None:
        """Update refresh rate"""
        self._database.write(
            "display",
            f"{display_key}_refresh_rate",
            self._display.refresh_rate(self._database, display_key),
        )

    async def update_resolution_horizontal(
        self,
        display_key: str,
    ) -> None:
        """Update resolution horizontal"""
        self._database.write(
            "display",
            f"{display_key}_resolution_horizontal",
            self._display.resolution_horizontal(self._database, display_key),
        )

    async def update_resolution_vertical(
        self,
        display_key: str,
    ) -> None:
        """Update resolution vertical"""
        self._database.write(
            "display",
            f"{display_key}_resolution_vertical",
            self._display.resolution_vertical(self._database, display_key),
        )

    async def update_all_data(self) -> None:
        """Update data"""

        # Clear table in case of hardware changes since last run
        self._database.clear_table("display")

        display_list = []
        for display_name in self._display.get_displays(self._database):
            display_key = make_key(display_name)
            display_list.append(display_key)
            await asyncio.gather(
                *[
                    self.update_name(display_key, display_name),
                    self.update_pixel_clock(display_key),
                    self.update_refresh_rate(display_key),
                    self.update_resolution_horizontal(display_key),
                    self.update_resolution_vertical(display_key),
                ]
            )
        self._database.write("display", "displays", display_list)



"""System Bridge: Memory"""
from typing import NamedTuple

from psutil import swap_memory, virtual_memory
from psutil._common import sswap
from systembridgeshared.base import Base


class Memory(Base):
    """Memory"""

    def swap(self) -> sswap:
        """Swap memory"""
        return swap_memory()

    def virtual(self) -> NamedTuple:  # svmem:
        """Virtual memory"""
        return virtual_memory()



"""System Bridge: Server Handler - Update"""
from sanic.request import Request
from sanic.response import HTTPResponse, json
from systembridgeshared.settings import Settings
from systembridgeshared.update import Update


async def handler_update(
    request: Request,
    _: Settings,
) -> HTTPResponse:
    """Handle the update request."""
    versions = Update().update(
        request.args.get("version"),
        wait=False,
    )
    return json(
        {
            "message": "Updating the application",
            "versions": versions,
        }
    )



"""System Bridge: Update Network"""
import asyncio

from systembridgeshared.database import Database

from systembridgebackend.modules.base import ModuleUpdateBase
from systembridgebackend.modules.network import Network


class NetworkUpdate(ModuleUpdateBase):
    """Network Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database, "network")
        self._network = Network()

    async def update_stats(self) -> None:
        """Update stats"""
        for key, value in self._network.stats().items():
            for subkey, subvalue in value._asdict().items():
                self._database.write(
                    "network", f"stat_{key.replace(' ', '')}_{subkey}", subvalue
                )

    async def update_io_counters(self) -> None:
        """Update IO counters"""
        for key, value in self._network.io_counters()._asdict().items():
            self._database.write("network", f"io_counters_{key}", value)

    async def update_all_data(self) -> None:
        """Update data"""
        await asyncio.gather(
            *[
                self.update_stats(),
                self.update_io_counters(),
            ]
        )



"""MDNS/Zeroconf Advertisement"""
from systembridgeshared.base import Base
from systembridgeshared.settings import SETTING_PORT_API, Settings
from zeroconf import InterfaceChoice, ServiceInfo, Zeroconf

from systembridgebackend.modules.system import System

ZEROCONF_TYPE = "_system-bridge._tcp.local."


class MDNSAdvertisement(Base):
    """MDNS/Zeroconf Advertisement"""

    def __init__(
        self,
        settings: Settings,
    ) -> None:
        """Initialize"""
        super().__init__()
        self._settings = settings

    def advertise_server(self) -> None:
        """Advertise server"""

        system = System()

        fqdn = system.fqdn()
        hostname = system.hostname()
        ip_address_4 = system.ip_address_4()
        mac_address = system.mac_address()
        port_api = self._settings.get(SETTING_PORT_API)
        system_id = system.uuid()

        if not port_api:
            raise ValueError("Port API not set")

        zeroconf = Zeroconf(
            interfaces=InterfaceChoice.All,
            unicast=True,
        )

        info = ServiceInfo(
            ZEROCONF_TYPE,
            name=f"{system_id}.{ZEROCONF_TYPE}",
            server=f"{system_id}.local.",
            parsed_addresses=[ip_address_4],
            port=int(port_api),
            properties={
                "address": f"http://{fqdn}:{port_api}",
                "fqdn": fqdn,
                "host": hostname,
                "ip": ip_address_4,
                "mac": mac_address,
                "port": port_api,
                "uuid": system_id,
                "version": system.version,
                "websocketAddress": f"ws://{fqdn}:{port_api}/api/websocket",
            },
        )

        self._logger.debug("Advertise: %s", info)

        zeroconf.register_service(info, allow_name_change=True)



"""System Bridge: CORS"""
from collections.abc import Iterable


def _add_cors_headers(response, methods: Iterable[str]) -> None:
    """Add CORS headers to a response."""
    allow_methods = list(set(methods))
    if "OPTIONS" not in allow_methods:
        allow_methods.append("OPTIONS")
    response.headers.extend(
        {
            "Access-Control-Allow-Methods": ",".join(allow_methods),
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Credentials": "true",
            "Access-Control-Allow-Headers": (
                "origin, content-type, accept, "
                "authorization, x-xsrf-token, x-request-id"
            ),
        }
    )


def add_cors_headers(request, response):
    """Add CORS headers to a response."""
    if request.method != "OPTIONS":
        _add_cors_headers(response, list(request.route.methods))



"""System Bridge: Battery"""
from __future__ import annotations

from typing import Optional

from plyer import battery
import psutil
from systembridgeshared.base import Base


class Battery(Base):
    """Battery"""

    def sensors(self) -> Optional[psutil._common.sfan]:
        """Get battery sensors"""
        if not hasattr(psutil, "sensors_battery"):
            return None
        return psutil.sensors_battery()  # type: ignore

    def status(self) -> dict:
        """Get battery status"""
        return battery.status



"""System Bridge: Update GPU"""
import asyncio

from systembridgeshared.common import make_key
from systembridgeshared.database import Database

from systembridgebackend.modules.base import ModuleUpdateBase
from systembridgebackend.modules.gpu import GPU


class GPUUpdate(ModuleUpdateBase):
    """GPU Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database, "gpu")
        self._gpu = GPU()

    async def update_name(
        self,
        gpu_key: str,
        gpu_name: str,
    ) -> None:
        """Update name"""
        self._database.write("gpu", f"{gpu_key}_name", gpu_name)

    async def update_core_clock(
        self,
        gpu_key: str,
    ) -> None:
        """Update core clock"""
        self._database.write(
            "gpu",
            f"{gpu_key}_core_clock",
            self._gpu.core_clock(self._database, gpu_key),
        )

    async def update_core_load(
        self,
        gpu_key: str,
    ) -> None:
        """Update core load"""
        self._database.write(
            "gpu", f"{gpu_key}_core_load", self._gpu.core_load(self._database, gpu_key)
        )

    async def update_fan_speed(
        self,
        gpu_key: str,
    ) -> None:
        """Update fan speed"""
        self._database.write(
            "gpu", f"{gpu_key}_fan_speed", self._gpu.fan_speed(self._database, gpu_key)
        )

    async def update_memory_clock(
        self,
        gpu_key: str,
    ) -> None:
        """Update memory clock"""
        self._database.write(
            "gpu",
            f"{gpu_key}_memory_clock",
            self._gpu.memory_clock(self._database, gpu_key),
        )

    async def update_memory_load(
        self,
        gpu_key: str,
    ) -> None:
        """Update memory load"""
        self._database.write(
            "gpu",
            f"{gpu_key}_memory_load",
            self._gpu.memory_load(self._database, gpu_key),
        )

    async def update_memory_free(
        self,
        gpu_key: str,
    ) -> None:
        """Update memory free"""
        self._database.write(
            "gpu",
            f"{gpu_key}_memory_free",
            self._gpu.memory_free(self._database, gpu_key),
        )

    async def update_memory_used(
        self,
        gpu_key: str,
    ) -> None:
        """Update memory used"""
        self._database.write(
            "gpu",
            f"{gpu_key}_memory_used",
            self._gpu.memory_used(self._database, gpu_key),
        )

    async def update_memory_total(
        self,
        gpu_key: str,
    ) -> None:
        """Update memory total"""
        self._database.write(
            "gpu",
            f"{gpu_key}_memory_total",
            self._gpu.memory_total(self._database, gpu_key),
        )

    async def update_power(
        self,
        gpu_key: str,
    ) -> None:
        """Update power"""
        self._database.write(
            "gpu", f"{gpu_key}_power", self._gpu.power(self._database, gpu_key)
        )

    async def update_temperature(
        self,
        gpu_key: str,
    ) -> None:
        """Update temperature"""
        self._database.write(
            "gpu",
            f"{gpu_key}_temperature",
            self._gpu.temperature(self._database, gpu_key),
        )

    async def update_all_data(self) -> None:
        """Update data"""

        # Clear table in case of hardware changes since last run
        self._database.clear_table("gpu")

        gpu_list = []
        for gpu_name in self._gpu.get_gpus(self._database):
            gpu_key = make_key(gpu_name)
            gpu_list.append(gpu_key)
            await asyncio.gather(
                *[
                    self.update_name(gpu_key, gpu_name),
                    self.update_core_clock(gpu_key),
                    self.update_core_load(gpu_key),
                    self.update_fan_speed(gpu_key),
                    self.update_memory_clock(gpu_key),
                    self.update_memory_load(gpu_key),
                    self.update_memory_free(gpu_key),
                    self.update_memory_used(gpu_key),
                    self.update_memory_total(gpu_key),
                    self.update_power(gpu_key),
                    self.update_temperature(gpu_key),
                ]
            )
        self._database.write("gpu", "gpus", gpu_list)



"""System Bridge: Update System"""
import asyncio

from systembridgeshared.database import Database

from systembridgebackend.modules.base import ModuleUpdateBase
from systembridgebackend.modules.system import System


class SystemUpdate(ModuleUpdateBase):
    """System Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database, "system")
        self._system = System()

    async def update_boot_time(self) -> None:
        """Update boot time"""
        self._database.write("system", "boot_time", self._system.boot_time())

    async def update_fqdn(self) -> None:
        """Update FQDN"""
        self._database.write("system", "fqdn", self._system.fqdn())

    async def update_hostname(self) -> None:
        """Update hostname"""
        self._database.write("system", "hostname", self._system.hostname())

    async def update_ip_address_4(self) -> None:
        """Update IP address 4"""
        self._database.write("system", "ip_address_4", self._system.ip_address_4())

    async def update_mac_address(self) -> None:
        """Update MAC address"""
        self._database.write("system", "mac_address", self._system.mac_address())

    async def update_platform(self) -> None:
        """Update platform"""
        self._database.write("system", "platform", self._system.platform())

    async def update_platform_version(self) -> None:
        """Update platform version"""
        self._database.write(
            "system", "platform_version", self._system.platform_version()
        )

    async def update_uptime(self) -> None:
        """Update uptime"""
        self._database.write("system", "uptime", self._system.uptime())

    async def update_users(self) -> None:
        """Update users"""
        for user in self._system.users():
            for key, value in user._asdict().items():
                self._database.write(
                    "system", f"user_{user.name.replace(' ','_').lower()}_{key}", value
                )

    async def update_uuid(self) -> None:
        """Update UUID"""
        self._database.write("system", "uuid", self._system.uuid())

    async def update_version(self) -> None:
        """Update version"""
        self._database.write("system", "version", self._system.version())

    async def update_version_latest(self) -> None:
        """Update latest version"""
        release = await self._system.version_latest()
        if release and release.tag_name:
            self._database.write(
                "system",
                "version_latest",
                release.tag_name.replace("v", "") if release is not None else None,
            )

    async def update_version_newer_available(self) -> None:
        """Update newer version available"""
        self._database.write(
            "system",
            "version_newer_available",
            self._system.version_newer_available(self._database),
        )

    async def update_all_data(self) -> None:
        """Update data"""
        await asyncio.gather(
            *[
                self.update_boot_time(),
                self.update_fqdn(),
                self.update_hostname(),
                self.update_ip_address_4(),
                self.update_mac_address(),
                self.update_platform(),
                self.update_platform_version(),
                self.update_uptime(),
                self.update_users(),
                self.update_uuid(),
                self.update_version(),
                self.update_version_latest(),
            ]
        )
        # Run after other version updates
        await self.update_version_newer_available()



"""System Bridge: Disk"""
from __future__ import annotations

from typing import NamedTuple, Optional

from psutil import disk_io_counters, disk_partitions, disk_usage
from psutil._common import sdiskio, sdiskpart
from systembridgeshared.base import Base


class Disk(Base):
    """Disk"""

    def io_counters(self) -> sdiskio:
        """Disk IO counters"""
        return disk_io_counters()

    def io_counters_per_disk(self) -> sdiskio:
        """Disk IO counters per disk"""
        return disk_io_counters(perdisk=True)

    def partitions(self) -> list[sdiskpart]:
        """Disk partitions"""
        return disk_partitions(all=True)

    def usage(self, path: str) -> Optional[NamedTuple]:
        """Disk usage"""
        try:
            return disk_usage(path)
        except PermissionError as error:
            self._logger.error("PermissionError: %s", error)
            return None
